from django.apps import apps
from django.conf import settings

from .schema import get_current_schema
from .utils import get_tenant_database_alias


class SyncRouter:
    """
    A router to control which applications will be synced depending on the schema we're syncing.
    """

    def app_in_list(self, app_label, app_list):
        app_config = apps.get_app_config(app_label)
        app_config_full_name = f"{app_config.__module__}.{app_config.__class__.__name__}"
        return (app_config.name in app_list) or (app_config_full_name in app_list)

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        current_schema = get_current_schema()
        if db != get_tenant_database_alias() or current_schema is None:
            return False
        app_list = []
        if current_schema.schema_name == "public":
            app_list = settings.TENANTS["public"]["APPS"]
        elif current_schema.schema_name in settings.TENANTS:
            app_list = settings.TENANTS[current_schema.schema_name]["APPS"]
        else:
            app_list = settings.TENANTS["default"]["APPS"]
        if not app_list:
            return None
        return self.app_in_list(app_label, app_list)



from importlib import import_module

from django.conf import settings

BASE_BACKEND = "django.db.backends.postgresql"
ORIGINAL_BACKEND = getattr(settings, "PGSCHEMAS_ORIGINAL_BACKEND", BASE_BACKEND)
EXTRA_SEARCH_PATHS = getattr(settings, "PGSCHEMAS_EXTRA_SEARCH_PATHS", [])

base_backend = import_module(BASE_BACKEND + ".base")
original_backend = import_module(ORIGINAL_BACKEND + ".base")



import re

from asgiref.sync import iscoroutinefunction, sync_to_async
from django.conf import settings
from django.http import Http404
from django.shortcuts import redirect
from django.urls import clear_url_caches, set_urlconf
from django.utils.decorators import sync_and_async_middleware

from .schema import SchemaDescriptor, activate, activate_public
from .urlresolvers import get_urlconf_from_schema
from .utils import get_domain_model, remove_www


def strip_tenant_from_path_factory(prefix):
    def strip_tenant_from_path(path):
        return re.sub(r"^/{}/".format(prefix), "/", path)

    return strip_tenant_from_path


@sync_and_async_middleware
def TenantMiddleware(get_response):
    """
    This middleware should be placed at the very top of the middleware stack.
    Selects the proper static/dynamic tenant using the request host.
    """

    def logic(request):
        hostname = remove_www(request.get_host().split(":")[0])

        activate_public()
        tenant = None

        # Checking for static tenants
        for schema, data in settings.TENANTS.items():
            if schema in ["public", "default"]:
                continue
            if hostname in data["DOMAINS"]:
                tenant = SchemaDescriptor.create(schema_name=schema, domain_url=hostname)
                break

        # Checking for dynamic tenants
        else:
            DomainModel = get_domain_model()
            prefix = request.path.split("/")[1]
            domain = None

            if DomainModel is not None:
                try:
                    domain = DomainModel.objects.select_related("tenant").get(
                        domain=hostname, folder=prefix
                    )
                except DomainModel.DoesNotExist:
                    try:
                        domain = DomainModel.objects.select_related("tenant").get(
                            domain=hostname, folder=""
                        )
                    except DomainModel.DoesNotExist:
                        pass

            if domain is not None:
                tenant = domain.tenant
                tenant.domain_url = hostname
                tenant.folder = None
                request.strip_tenant_from_path = lambda x: x
                if prefix and domain.folder == prefix:
                    tenant.folder = prefix
                    request.strip_tenant_from_path = strip_tenant_from_path_factory(prefix)
                    clear_url_caches()  # Required to remove previous tenant prefix from cache (#8)
                if domain.redirect_to_primary:
                    primary_domain = tenant.domains.get(is_primary=True)
                    path = request.strip_tenant_from_path(request.path)
                    return redirect(primary_domain.absolute_url(path), permanent=True)

        # Checking fallback domains
        if not tenant:
            for schema, data in settings.TENANTS.items():
                if schema in ["public", "default"]:
                    continue
                if hostname in data.get("FALLBACK_DOMAINS", []):
                    tenant = SchemaDescriptor.create(schema_name=schema, domain_url=hostname)
                    break

        # No tenant found from domain / folder
        if not tenant:
            raise Http404("No tenant for hostname '%s'" % hostname)

        request.tenant = tenant
        urlconf = get_urlconf_from_schema(tenant)
        request.urlconf = urlconf
        set_urlconf(urlconf)

        activate(tenant)

    if iscoroutinefunction(get_response):
        async_logic = sync_to_async(logic)

        async def middleware(request):
            if response := await async_logic(request):
                return response

            return await get_response(request)

    else:

        def middleware(request):
            if response := logic(request):
                return response

            return get_response(request)

    return middleware



from django.test import Client, RequestFactory


def get_domain(tenant):
    if tenant.is_dynamic:
        return tenant.get_primary_domain().domain
    return tenant.domain_url or tenant.schema_name


class TenantRequestFactory(RequestFactory):
    def __init__(self, tenant, **defaults):
        super().__init__(**defaults)
        self.tenant = tenant

    def get(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().get(path, data, **extra)

    def post(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().post(path, data, **extra)

    def patch(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().patch(path, data, **extra)

    def put(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().put(path, data, **extra)

    def delete(self, path, data=None, content_type="application/octet-stream", **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().delete(path, data, **extra)


class TenantClient(Client):
    def __init__(self, tenant, enforce_csrf_checks=False, **defaults):
        super().__init__(enforce_csrf_checks, **defaults)
        self.tenant = tenant

    def get(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().get(path, data, **extra)

    def post(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().post(path, data, **extra)

    def patch(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().patch(path, data, **extra)

    def put(self, path, data=None, **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().put(path, data, **extra)

    def delete(self, path, data=None, content_type="application/octet-stream", **extra):
        if "HTTP_HOST" not in extra:
            extra["HTTP_HOST"] = get_domain(self.tenant)
        return super().delete(path, data, **extra)



import functools
import multiprocessing

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError, OutputWrapper
from django.db import connection, connections, transaction

from ...schema import SchemaDescriptor, activate
from ...utils import get_clone_reference, get_tenant_model


def run_on_schema(
    schema_name,
    executor_codename,
    command,
    function_name=None,
    args=None,
    kwargs=None,
    pass_schema_in_kwargs=False,
    fork_db=False,
):
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    if not isinstance(command, BaseCommand):
        # Parallel executor needs to pass command 'type' instead of 'instance'
        # Therefore, no customizations for the command can be done, nor using custom stdout, stderr
        command = command()

    command.stdout = kwargs.pop("stdout", command.stdout)
    if not isinstance(command.stdout, OutputWrapper):
        command.stdout = OutputWrapper(command.stdout)

    command.stderr = kwargs.pop("stderr", command.stderr)
    if not isinstance(command.stderr, OutputWrapper):
        command.stderr = OutputWrapper(command.stderr)

    # Since we are prepending every output with the schema_name and executor, we need to determine
    # whether we need to do so based on the last ending used to write. If the last write didn't end
    # in '\n' then we don't do the prefixing in order to keep the output looking good.
    class StyleFunc:
        last_message = None

        def __call__(self, message):
            last_message = self.last_message
            self.last_message = message
            if last_message is None or last_message.endswith("\n"):
                return "[%s:%s] %s" % (
                    command.style.NOTICE(executor_codename),
                    command.style.NOTICE(schema_name),
                    message,
                )
            return message

    command.stdout.style_func = StyleFunc()
    command.stderr.style_func = StyleFunc()

    if fork_db:
        connections.close_all()

    if schema_name in settings.TENANTS:
        domains = settings.TENANTS[schema_name].get("DOMAINS", [])
        schema = SchemaDescriptor.create(
            schema_name=schema_name, domain_url=domains[0] if domains else None
        )
    elif schema_name == get_clone_reference():
        schema = SchemaDescriptor.create(schema_name=schema_name)
    elif (TenantModel := get_tenant_model()) is not None:
        schema = TenantModel.objects.get(schema_name=schema_name)
    else:
        raise CommandError(f"Unable to find schema {schema_name}!")

    activate(schema)

    if pass_schema_in_kwargs:
        kwargs.update({"schema_name": schema_name})

    if function_name == "special:call_command":
        call_command(command, *args, **kwargs)
    elif function_name == "special:run_from_argv":
        command.run_from_argv(args)
    else:
        getattr(command, function_name)(*args, **kwargs)

    if fork_db:
        transaction.commit()
        connection.close()

    return schema_name


def sequential(
    schemas, command, function_name, args=None, kwargs=None, pass_schema_in_kwargs=False
):
    runner = functools.partial(
        run_on_schema,
        executor_codename="sequential",
        command=command,
        function_name=function_name,
        args=args,
        kwargs=kwargs,
        pass_schema_in_kwargs=pass_schema_in_kwargs,
        fork_db=False,
    )
    for schema in schemas:
        runner(schema)
    return schemas


def parallel(schemas, command, function_name, args=None, kwargs=None, pass_schema_in_kwargs=False):
    processes = getattr(settings, "PGSCHEMAS_PARALLEL_MAX_PROCESSES", None)
    pool = multiprocessing.Pool(processes=processes)
    runner = functools.partial(
        run_on_schema,
        executor_codename="parallel",
        command=type(command),  # Can't pass streams to children processes
        function_name=function_name,
        args=args,
        kwargs=kwargs,
        pass_schema_in_kwargs=pass_schema_in_kwargs,
        fork_db=True,
    )
    return pool.map(runner, schemas)



from channels.auth import AuthMiddleware, CookieMiddleware, SessionMiddleware, _get_user_session_key
from channels.db import database_sync_to_async
from django.conf import settings
from django.contrib.auth import BACKEND_SESSION_KEY, HASH_SESSION_KEY, load_backend
from django.contrib.auth.models import AnonymousUser
from django.utils.crypto import constant_time_compare


@database_sync_to_async
def get_user(scope):
    """
    Return the user model instance associated with the given scope.
    If no user is retrieved, return an instance of `AnonymousUser`.
    """
    if "session" not in scope:
        raise ValueError(
            "Cannot find session in scope. You should wrap your consumer in SessionMiddleware."
        )
    user = None
    session = scope["session"]
    with scope["tenant"]:
        try:
            user_id = _get_user_session_key(session)
            backend_path = session[BACKEND_SESSION_KEY]
        except KeyError:
            pass
        else:
            if backend_path in settings.AUTHENTICATION_BACKENDS:
                backend = load_backend(backend_path)
                user = backend.get_user(user_id)
                # Verify the session
                if hasattr(user, "get_session_auth_hash"):
                    session_hash = session.get(HASH_SESSION_KEY)
                    session_hash_verified = session_hash and constant_time_compare(
                        session_hash, user.get_session_auth_hash()
                    )
                    if not session_hash_verified:
                        session.flush()
                        user = None
    return user or AnonymousUser()


class TenantAuthMiddleware(AuthMiddleware):
    async def resolve_scope(self, scope):
        scope["user"]._wrapped = await get_user(scope)


def TenantAuthMiddlewareStack(inner):
    return CookieMiddleware(SessionMiddleware(TenantAuthMiddleware(inner)))



from django.db.models.signals import pre_delete
from django.dispatch import Signal, receiver

from .utils import get_tenant_model, schema_exists

schema_activate = Signal()
schema_activate.__doc__ = "Sent after a schema has been activated"

dynamic_tenant_needs_sync = Signal()
dynamic_tenant_needs_sync.__doc__ = "Sent when a schema from a dynamic tenant needs to be synced"

dynamic_tenant_post_sync = Signal()
dynamic_tenant_post_sync.__doc__ = (
    "Sent after a tenant has been saved, its schema created and synced"
)

dynamic_tenant_pre_drop = Signal()
dynamic_tenant_pre_drop.__doc__ = "Sent when a schema from a dynamic tenant is about to be dropped"


@receiver(pre_delete)
def tenant_delete_callback(sender, instance, **kwargs):
    TenantModel = get_tenant_model()
    if TenantModel is None:
        return
    if not isinstance(instance, TenantModel):
        return
    if instance.auto_drop_schema and schema_exists(instance.schema_name):
        dynamic_tenant_pre_drop.send(sender=TenantModel, tenant=instance.serializable_fields())
        instance.drop_schema()



from . import TenantCommand


class Command(TenantCommand):
    help = "Displays which schemas would be used based on the passed schema selectors"

    def handle_tenant(self, tenant, *args, **options):
        if options["verbosity"] >= 1:
            self.stdout.write(
                str(tenant.get_primary_domain())
                if tenant.is_dynamic
                else tenant.domain_url or tenant.schema_name
            )



from django.core.exceptions import ImproperlyConfigured
from django.db.utils import DatabaseError

from ..schema import get_current_schema, get_default_schema
from ..utils import check_schema_name, get_limit_set_calls
from .introspection import DatabaseSchemaIntrospection
from .settings import EXTRA_SEARCH_PATHS, original_backend

try:
    try:
        import psycopg as _psycopg
    except ImportError:
        import psycopg2 as _psycopg
except ImportError:
    raise ImproperlyConfigured("Error loading psycopg2 or psycopg module")

IntegrityError = _psycopg.IntegrityError


def get_search_path(schema=None):
    if schema is None:
        schema = get_default_schema()

    search_path = ["public"] if schema.schema_name == "public" else [schema.schema_name, "public"]
    search_path.extend(EXTRA_SEARCH_PATHS)

    for part in search_path:
        check_schema_name(part)

    return ", ".join(search_path)


class DatabaseWrapper(original_backend.DatabaseWrapper):
    def __init__(self, *args, **kwargs):
        self._search_path = None
        self._setting_search_path = False
        super().__init__(*args, **kwargs)

        # Patched version of DatabaseIntrospection that only returns the table list for the currently selected schema
        self.introspection = DatabaseSchemaIntrospection(self)

    def close(self):
        self._search_path = None
        self._setting_search_path = False
        super().close()

    def _handle_search_path(self, cursor=None):
        search_path_for_current_schema = get_search_path(get_current_schema())

        skip = self._setting_search_path or (
            self._search_path == search_path_for_current_schema and get_limit_set_calls()
        )

        if not skip:
            self._setting_search_path = True
            cursor_for_search_path = self.connection.cursor() if cursor is None else cursor

            try:
                cursor_for_search_path.execute(
                    f"SET search_path = {search_path_for_current_schema}"
                )
            except (DatabaseError, _psycopg.InternalError):
                self._search_path = None
            else:
                self._search_path = search_path_for_current_schema
            finally:
                self._setting_search_path = False

            if cursor is None:
                cursor_for_search_path.close()

    def _cursor(self, name=None):
        cursor = super()._cursor(name=name)

        cursor_for_search_path = cursor if name is None else None  # Named cursors cannot be reused
        self._handle_search_path(cursor_for_search_path)

        return cursor



import re
import sys

from django.conf import settings
from django.urls import URLResolver

from .schema import SchemaDescriptor, get_current_schema


class TenantPrefixPattern:
    converters = {}

    @property
    def tenant_prefix(self):
        current_schema = get_current_schema()
        return f"{current_schema.folder}/" if current_schema.folder else "/"

    @property
    def regex(self):
        # This is only used by reverse() and cached in _reverse_dict.
        return re.compile(self.tenant_prefix)

    def match(self, path):
        tenant_prefix = self.tenant_prefix
        if path.startswith(tenant_prefix):
            return path[len(tenant_prefix) :], (), {}
        return None

    def check(self):
        return []

    def describe(self):
        return f"'{self}'"

    def __str__(self):
        return self.tenant_prefix


def tenant_patterns(*urls):
    """
    Add the tenant prefix to every URL pattern within this function.
    This may only be used in the root URLconf, not in an included URLconf.
    """
    return [URLResolver(TenantPrefixPattern(), list(urls))]


def get_dynamic_tenant_prefixed_urlconf(urlconf, dynamic_path):
    """
    Generates a new URLConf module with all patterns prefixed with tenant.
    """
    from types import ModuleType

    from django.utils.module_loading import import_string

    class LazyURLConfModule(ModuleType):
        def __getattr__(self, attr):
            imported = import_string(f"{urlconf}.{attr}")
            if attr == "urlpatterns":
                return tenant_patterns(*imported)
            return imported

    return LazyURLConfModule(dynamic_path)


def get_urlconf_from_schema(schema):
    """
    Returns the proper URLConf depending on the schema.
    The schema must come with ``domain_url`` and ``folder`` members set.
    """
    assert isinstance(schema, SchemaDescriptor)

    if not schema.domain_url:
        return None

    # Checking for static tenants
    if not schema.is_dynamic:
        for schema_name, data in settings.TENANTS.items():
            if schema_name in ["public", "default"]:
                continue
            if schema.domain_url in data["DOMAINS"]:
                return data["URLCONF"]
            if schema.domain_url in data.get("FALLBACK_DOMAINS", []):
                return data["URLCONF"]
        return None

    # Checking for dynamic tenants
    urlconf = settings.TENANTS["default"]["URLCONF"]
    if schema.folder:
        dynamic_path = urlconf + "_dynamically_tenant_prefixed"
        if not sys.modules.get(dynamic_path):
            sys.modules[dynamic_path] = get_dynamic_tenant_prefixed_urlconf(urlconf, dynamic_path)
        urlconf = dynamic_path

    return urlconf



import os
import re
from typing import Optional

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.db import DEFAULT_DB_ALIAS, ProgrammingError, connection, transaction
from django.db.models import Model


def get_tenant_model(require_ready: bool = True) -> Optional[Model]:
    "Returns the tenant model."
    if "default" not in settings.TENANTS:
        return None
    return apps.get_model(settings.TENANTS["default"]["TENANT_MODEL"], require_ready=require_ready)


def get_domain_model(require_ready: bool = True) -> Optional[Model]:
    "Returns the domain model."
    if "default" not in settings.TENANTS:
        return None
    return apps.get_model(settings.TENANTS["default"]["DOMAIN_MODEL"], require_ready=require_ready)


def get_tenant_database_alias() -> str:
    return getattr(settings, "PGSCHEMAS_TENANT_DB_ALIAS", DEFAULT_DB_ALIAS)


def get_limit_set_calls() -> bool:
    return getattr(settings, "PGSCHEMAS_LIMIT_SET_CALLS", False)


def get_clone_reference() -> Optional[str]:
    if "default" not in settings.TENANTS:
        return None
    return settings.TENANTS["default"].get("CLONE_REFERENCE", None)


def is_valid_identifier(identifier: str) -> bool:
    "Checks the validity of identifier."
    SQL_IDENTIFIER_RE = re.compile(r"^[_a-zA-Z][_a-zA-Z0-9]{,62}$")
    return bool(SQL_IDENTIFIER_RE.match(identifier))


def is_valid_schema_name(name: str) -> bool:
    "Checks the validity of a schema name."
    SQL_SCHEMA_NAME_RESERVED_RE = re.compile(r"^pg_", re.IGNORECASE)
    return is_valid_identifier(name) and not SQL_SCHEMA_NAME_RESERVED_RE.match(name)


def check_schema_name(name: str):
    """
    Checks schema name and raises ``ValidationError`` if ``name`` is not a
    valid identifier.
    """
    if not is_valid_schema_name(name):
        raise ValidationError("Invalid string used for the schema name.")


def remove_www(hostname: str) -> str:
    """
    Removes ``www``. from the beginning of the address. Only for
    routing purposes. ``www.test.com/login/`` and ``test.com/login/`` should
    find the same tenant.
    """
    if hostname.startswith("www."):
        return hostname[4:]
    return hostname


def django_is_in_test_mode() -> bool:
    """
    I know this is very ugly! I'm looking for more elegant solutions.
    See: http://stackoverflow.com/questions/6957016/detect-django-testing-mode
    """
    from django.core import mail

    return hasattr(mail, "outbox")


def run_in_public_schema(func):
    "Decorator that makes decorated function to be run in the public schema."

    def wrapper(*args, **kwargs):
        from .schema import SchemaDescriptor

        with SchemaDescriptor.create(schema_name="public"):
            return func(*args, **kwargs)

    return wrapper


def schema_exists(schema_name: str) -> bool:
    "Checks if a schema exists in database."
    sql = """
    SELECT EXISTS(
        SELECT 1
        FROM pg_catalog.pg_namespace
        WHERE LOWER(nspname) = LOWER(%s)
    )
    """
    cursor = connection.cursor()
    cursor.execute(sql, (schema_name,))
    row = cursor.fetchone()
    if row:
        exists = row[0]
    else:  # pragma: no cover
        exists = False
    cursor.close()
    return exists


@run_in_public_schema
def dynamic_models_exist() -> bool:
    "Checks if tenant model and domain model have been synced."
    sql = """
    SELECT count(*)
    FROM   information_schema.tables
    WHERE  table_schema = 'public'
    AND    table_name in ('%s', '%s');
    """
    TenantModel = get_tenant_model()
    DomainModel = get_domain_model()

    if TenantModel is None or DomainModel is None:
        return False

    cursor = connection.cursor()
    cursor.execute(sql % (TenantModel._meta.db_table, DomainModel._meta.db_table))
    value = cursor.fetchone() == (2,)
    cursor.close()
    return value


@run_in_public_schema
def create_schema(
    schema_name: str,
    check_if_exists: bool = False,
    sync_schema: bool = True,
    verbosity: int = 1,
) -> bool:
    """
    Creates the schema ``schema_name``. Optionally checks if the schema already
    exists before creating it. Returns ``True`` if the schema was created,
    ``False`` otherwise.
    """
    check_schema_name(schema_name)
    if check_if_exists and schema_exists(schema_name):
        return False
    cursor = connection.cursor()
    cursor.execute("CREATE SCHEMA %s" % schema_name)
    cursor.close()
    if sync_schema:
        call_command("migrateschema", schemas=[schema_name], verbosity=verbosity)
    return True


@run_in_public_schema
def drop_schema(schema_name: str, check_if_exists: bool = True, verbosity: int = 1) -> bool:
    """
    Drops the schema. Optionally checks if the schema already exists before
    dropping it.
    """
    if check_if_exists and not schema_exists(schema_name):
        return False
    cursor = connection.cursor()
    cursor.execute("DROP SCHEMA %s CASCADE" % schema_name)
    cursor.close()
    return True


class DryRunException(Exception):
    pass


def _create_clone_schema_function():
    """
    Creates a postgres function `clone_schema` that copies a schema and its
    contents. Will replace any existing `clone_schema` functions owned by the
    `postgres` superuser.
    """
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "clone_schema.sql")) as f:
        CLONE_SCHEMA_FUNCTION = (
            f.read()
            .replace("RAISE NOTICE ' source schema", "RAISE EXCEPTION ' source schema")
            .replace("RAISE NOTICE ' dest schema", "RAISE EXCEPTION ' dest schema")
        )

    cursor = connection.cursor()
    cursor.execute(CLONE_SCHEMA_FUNCTION)
    cursor.close()


@run_in_public_schema
def clone_schema(base_schema_name: str, new_schema_name: str, dry_run: bool = False):
    """
    Creates a new schema ``new_schema_name`` as a clone of an existing schema
    ``base_schema_name``.
    """
    check_schema_name(new_schema_name)
    cursor = connection.cursor()

    # check if the clone_schema function already exists in the db
    try:
        cursor.execute(
            "SELECT 'public.clone_schema(text, text, public.cloneparms[])'::regprocedure"
        )
    except ProgrammingError:  # pragma: no cover
        _create_clone_schema_function()
        transaction.commit()

    try:
        with transaction.atomic():
            cursor.callproc("clone_schema", [base_schema_name, new_schema_name, "DATA"])
            cursor.close()
            if dry_run:
                raise DryRunException
    except DryRunException:
        cursor.close()


def create_or_clone_schema(schema_name: str, sync_schema: bool = True, verbosity: int = 1) -> bool:
    """
    Creates the schema ``schema_name``. Optionally checks if the schema already
    exists before creating it. Returns ``True`` if the schema was created,
    ``False`` otherwise.
    """
    check_schema_name(schema_name)
    if schema_exists(schema_name):
        return False
    clone_reference = get_clone_reference()
    if (
        clone_reference and schema_exists(clone_reference) and not django_is_in_test_mode()
    ):  # pragma: no cover
        clone_schema(clone_reference, schema_name)
        return True
    return create_schema(schema_name, sync_schema=sync_schema, verbosity=verbosity)



from typing import Optional

from asgiref.local import Local

from .signals import schema_activate

_active = Local()


def get_default_schema() -> "SchemaDescriptor":
    return SchemaDescriptor.create("public")


def get_current_schema() -> "SchemaDescriptor":
    current_schema = getattr(_active, "value", None)
    return current_schema or get_default_schema()


def activate(schema: "SchemaDescriptor"):
    if not isinstance(schema, SchemaDescriptor):
        raise RuntimeError("'activate' must be called with a SchemaDescriptor descendant")

    _active.value = schema
    schema_activate.send(sender=SchemaDescriptor, schema=schema)


def deactivate():
    if hasattr(_active, "value"):
        del _active.value
    schema_activate.send(sender=SchemaDescriptor, schema=SchemaDescriptor.create("public"))


activate_public = deactivate


class SchemaDescriptor:
    schema_name = None
    domain_url = None
    folder = None

    is_dynamic = False

    @staticmethod
    def create(schema_name: str, domain_url: Optional[str] = None, folder: Optional[str] = None):
        schema = SchemaDescriptor()
        schema.schema_name = schema_name
        schema.domain_url = domain_url
        schema.folder = folder
        return schema

    def __enter__(self):
        self.previous_schema = get_current_schema()
        activate(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        previous_schema = getattr(self, "previous_schema", None)
        activate(previous_schema) if previous_schema else deactivate()



from django.core.checks import Tags, run_checks
from django.core.management.base import BaseCommand, CommandError

from ...utils import create_schema, drop_schema, get_clone_reference


class Command(BaseCommand):
    help = "Creates the reference schema for faster dynamic tenant creation"

    def _run_checks(self, **kwargs):  # pragma: no cover
        issues = run_checks(tags=[Tags.database])
        issues.extend(super()._run_checks(**kwargs))
        return issues

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--recreate",
            action="store_true",
            dest="recreate",
            help="Recreate reference schema.",
        )

    def handle(self, *args, **options):
        clone_reference = get_clone_reference()
        if not clone_reference:
            raise CommandError("There is no reference schema configured.")
        if options.get("recreate", False):
            drop_schema(clone_reference, check_if_exists=True, verbosity=options["verbosity"])
            if options["verbosity"] >= 1:
                self.stdout.write("Destroyed existing reference schema.")
        created = create_schema(
            clone_reference, check_if_exists=True, verbosity=options["verbosity"]
        )
        if options["verbosity"] >= 1:
            if created:
                self.stdout.write("Reference schema successfully created!")
            else:
                self.stdout.write("Reference schema already exists.")
                self.stdout.write(
                    self.style.WARNING(
                        "Run this command again with --recreate if you want to recreate the reference schema."
                    )
                )



from typing import Any, Optional

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sessions.base_session import AbstractBaseSession
from django.core import checks
from django.core.exceptions import ImproperlyConfigured
from django.db.utils import ProgrammingError
from django.utils.module_loading import import_module

from .utils import get_clone_reference, get_domain_model, get_tenant_model


def get_tenant_app() -> Optional[str]:
    TenantModel = get_tenant_model(require_ready=False)
    if TenantModel is None:
        return None
    return TenantModel._meta.app_config.name


def get_domain_app() -> Optional[str]:
    DomainModel = get_domain_model(require_ready=False)
    if DomainModel is None:
        return None
    return DomainModel._meta.app_config.name


def get_user_app() -> Optional[str]:
    try:
        return get_user_model()._meta.app_config.name
    except ImproperlyConfigured:
        return None


def get_session_app() -> Optional[str]:
    engine = import_module(settings.SESSION_ENGINE)
    store = engine.SessionStore
    if hasattr(store, "get_model_class"):
        session_model = store.get_model_class()
        if issubclass(session_model, AbstractBaseSession):
            return session_model._meta.app_config.name
    return None


@checks.register()
def check_principal_apps(app_configs: Any, **kwargs: Any) -> None:
    errors = []
    tenant_app = get_tenant_app()
    domain_app = get_domain_app()

    if tenant_app is None or domain_app is None:
        return []

    if tenant_app not in settings.TENANTS["public"].get("APPS", []):
        errors.append(
            checks.Error(
                "Your tenant app '%s' must be on the 'public' schema." % tenant_app,
                id="pgschemas.W001",
            )
        )
    if domain_app not in settings.TENANTS["public"].get("APPS", []):
        errors.append(
            checks.Error(
                "Your domain app '%s' must be on the 'public' schema." % domain_app,
                id="pgschemas.W001",
            )
        )

    for schema in settings.TENANTS:
        schema_apps = settings.TENANTS[schema].get("APPS", [])
        if schema == "public":
            continue
        if tenant_app in schema_apps:
            errors.append(
                checks.Error(
                    "Your tenant app '%s' in TENANTS['%s']['APPS'] must be on the 'public' schema only."
                    % (tenant_app, schema),
                    id="pgschemas.W001",
                )
            )
        if domain_app in schema_apps:
            errors.append(
                checks.Error(
                    "Your domain app '%s' in TENANTS['%s']['APPS'] must be on the 'public' schema only."
                    % (domain_app, schema),
                    id="pgschemas.W001",
                )
            )

    return errors


@checks.register()
def check_other_apps(app_configs: Any, **kwargs: Any) -> None:
    errors = []
    user_app = get_user_app()
    session_app = get_session_app()

    if "django.contrib.contenttypes" in settings.TENANTS.get("default", {}).get("APPS", []):
        errors.append(
            checks.Warning(
                "'django.contrib.contenttypes' in TENANTS['default']['APPS'] must be on 'public' schema only.",
                id="pgschemas.W002",
            )
        )

    for schema in settings.TENANTS:
        schema_apps = settings.TENANTS[schema].get("APPS", [])
        if schema not in ["public", "default"]:
            if "django.contrib.contenttypes" in schema_apps:
                errors.append(
                    checks.Warning(
                        "'django.contrib.contenttypes' in TENANTS['%s']['APPS'] must be on 'public' schema only."
                        % schema,
                        id="pgschemas.W002",
                    )
                )
        if user_app and session_app:
            if session_app in schema_apps and user_app not in schema_apps:
                errors.append(
                    checks.Warning(
                        "'%s' must be together with '%s' in TENANTS['%s']['APPS']."
                        % (user_app, session_app, schema),
                        id="pgschemas.W003",
                    )
                )
            elif (
                user_app in schema_apps
                and session_app not in schema_apps
                and session_app in settings.INSTALLED_APPS
            ):
                errors.append(
                    checks.Warning(
                        "'%s' must be together with '%s' in TENANTS['%s']['APPS']."
                        % (session_app, user_app, schema),
                        id="pgschemas.W003",
                    )
                )

    return errors


@checks.register(checks.Tags.database)
def check_schema_names(app_configs: Any, **kwargs: Any) -> None:
    errors = []
    static_names = set(settings.TENANTS.keys())
    clone_reference = get_clone_reference()
    TenantModel = get_tenant_model()

    if TenantModel is None:
        return []

    if clone_reference:
        static_names.add(clone_reference)
    try:
        dynamic_names = set(TenantModel.objects.values_list("schema_name", flat=True))
    except ProgrammingError:
        # This happens on the first run of migrate, with empty database.
        # It can also happen when the tenant model contains unapplied migrations that break.
        dynamic_names = set()

    intersection = static_names & dynamic_names

    if intersection:
        errors.append(
            checks.Critical(
                "Name clash found between static and dynamic tenants: %s" % intersection,
                id="pgschemas.W004",
            )
        )

    return errors



from .schema import (
    SchemaDescriptor,
    activate,
    activate_public,
    deactivate,
    get_current_schema,
)

__all__ = [
    "SchemaDescriptor",
    "activate",
    "activate_public",
    "deactivate",
    "get_current_schema",
]
